//#include <stdio.h>
#pragma import(__use_no_semihosting_swi)
#include <stdlib.h>
// #include "sdk_config.h"
#include "nrfx_config.h"
#include "nrfx_spim.h"
#include "nrfx_timer.h"
#include "nrf_gpio.h"
#include "nrfx_gpiote.h"
#include "nrfx_ppi.h"
#include "nrfx_coredep.h" // for delay function
#include "app_error.h"
#include <math.h>

#include "nrf_delay.h"

// Define the pins used for testing on the nRF52840_DK board
#include "pca10056.h"

// Translation macro for error handling
#define NFRX_APP_ERROR_CHECK(err_code) APP_ERROR_CHECK(err_code)

// Define the pins used on the Headphones board
#define AIN0_PIN     NRF_GPIO_PIN_MAP(0,2)  // P0.02 = SIO_02
#define AIN1_PIN     NRF_GPIO_PIN_MAP(0,3)  // P0.03 = SIO_03
#define SPI_MISO_PIN NRF_GPIO_PIN_MAP(0,4)  // P0.04 = SIO_04
#define UART_RTS_PIN NRF_GPIO_PIN_MAP(0,5)  // P0.05 = SIO_05
#define UART_TX_PIN  NRF_GPIO_PIN_MAP(0,6)  // P0.06 = SIO_06
#define UART_CTS_PIN NRF_GPIO_PIN_MAP(0,7)  // P0.07 = SIO_07
#define UART_RX_PIN  NRF_GPIO_PIN_MAP(0,8)  // P0.08 = SIO_08
#define GPOUT_PIN    NRF_GPIO_PIN_MAP(0,12) // P0.12 = SIO_12
#define SW_PIN       NRF_GPIO_PIN_MAP(0,13) // P0.13 = SIO_13
#define B_PIN        NRF_GPIO_PIN_MAP(0,14) // P0.14 = SIO_16
#define A_PIN        NRF_GPIO_PIN_MAP(0,16) // P0.16 = SIO_14
#define RESET_PIN    NRF_GPIO_PIN_MAP(0,18) // P0.18 = SIO_18
#define POWER_EN_PIN NRF_GPIO_PIN_MAP(0,24) // P0.24 = SIO_24
#define I2C_SDA_PIN  NRF_GPIO_PIN_MAP(0,26) // P0.26 = SIO_26
#define I2C_SCL_PIN  NRF_GPIO_PIN_MAP(0,27) // P0.27 = SIO_27
#define AIN4_PIN     NRF_GPIO_PIN_MAP(0,28) // P0.28 = SIO_28
#define AIN5_PIN     NRF_GPIO_PIN_MAP(0,29) // P0.29 = SIO_29
#define AIN6_PIN     NRF_GPIO_PIN_MAP(0,30) // P0.30 = SIO_30
#define AIN7_PIN     NRF_GPIO_PIN_MAP(0,31) // P0.31 = SIO_31

#define KG_PIN       NRF_GPIO_PIN_MAP(1,2)  // P1.02 = SIO_34
#define KR_PIN       NRF_GPIO_PIN_MAP(1,4)  // P1.04 = SIO_36
#define SW1_PIN      NRF_GPIO_PIN_MAP(1,5)  // P1.05 = SIO_37
#define SPI_MOSI_PIN NRF_GPIO_PIN_MAP(1,8)  // P1.08 = SIO_40
#define SPI_SCK_PIN  NRF_GPIO_PIN_MAP(1,9)  // P1.09 = SIO_41
#define SW3_PIN      NRF_GPIO_PIN_MAP(1,10) // P1.10 = SIO_42
#define SW2_PIN      NRF_GPIO_PIN_MAP(1,11) // P1.11 = SIO_43
#define SPI_CSN_PIN  NRF_GPIO_PIN_MAP(1,12) // P1.12 = SIO_44
#define SPI_LDAC_PIN NRF_GPIO_PIN_MAP(1,13) // P1.13 = SIO_45
#define JACK_EN_PIN  NRF_GPIO_PIN_MAP(1,14) // P1.14 = SIO_46
#define ISET2_PIN    NRF_GPIO_PIN_MAP(1,15) // P1.15 = SIO_47

#define DAC_A    0x0000 // DAC Channel Selection (0=>A, 1=>B)
#define DAC_B    0x8000 // DAC Channel Selection (0=>A, 1=>B)
#define DAC_GAx1 0x2000 // DAC Gain Selection (0=>2x, 1=>1x)
#define DAC_GAx2 0x0000 // DAC Gain Selection (0=>2x, 1=>1x)
#define DAC_RUN  0x1000 // DAC Shutdown (0=>SHUTDOWN, 1=>RUNNING)
#define DAC_SHDN 0x0000 // DAC Shutdown (0=>SHUTDOWN, 1=>RUNNING)

#define AngleStep 0.0098174770424681038701957605727484

// EasyDMA configuration for the SPI DAC
#define SPI_INSTANCE 3 // SPI instance index.
#define BUFFER_SIZE 2 // There are 2 bytes in each SPI transfer to the DAC
#define NUMBER_OF_BUFFERS 960 // There are 960 2-byte transfers to the DAC every 30ms
static const nrfx_spim_t spim3 = NRFX_SPIM_INSTANCE(SPI_INSTANCE);  // SPI instance.
static volatile bool WriterListTransmitting = 0;
  
typedef struct ArrayList
{
  volatile uint8_t buffer[BUFFER_SIZE];
} ArrayList_type;
static volatile ArrayList_type WriterList0[NUMBER_OF_BUFFERS] __attribute__((at(0x20005000))); //NB: __at__ is not supported in SEGGER due to GCC restrictions.
static volatile ArrayList_type WriterList1[NUMBER_OF_BUFFERS] __attribute__((at(0x20006000))); //NB: __at__ is not supported in SEGGER due to GCC restrictions.

// Declare the Timer instances (NB: Timer 0 is used by the soft device)
static const nrfx_timer_t m_timer1 = NRFX_TIMER_INSTANCE(1);
static const nrfx_timer_t m_timer2 = NRFX_TIMER_INSTANCE(2);

// Declare the PPI instances
nrf_ppi_channel_t ppi_channel0;
nrf_ppi_channel_t ppi_channel1;
nrf_ppi_channel_t ppi_channel2;
nrf_ppi_channel_t ppi_channel3;
nrf_ppi_channel_t ppi_channel4;

void GPIO_init(void)
{
  nrfx_err_t err_code = NRFX_SUCCESS;
  // Initialise the GPIOTE driver if it hasn't already been initilaised
  if(!nrfx_gpiote_is_init())
  {
    err_code = nrfx_gpiote_init();
    NFRX_APP_ERROR_CHECK(err_code);
  }

  // set the IO direction and IO type (digital/analog) for every used pin on the IC.

  // Digital Outputs
  nrf_gpio_cfg_output(SPI_SCK_PIN);
  nrf_gpio_cfg_output(SPI_MOSI_PIN);
  nrf_gpio_cfg_output(SPI_CSN_PIN);
  nrf_gpio_cfg_output(SPI_LDAC_PIN);
  nrf_gpio_cfg_output(ISET2_PIN);
  nrf_gpio_cfg_output(KR_PIN);
  nrf_gpio_cfg_output(KG_PIN);
  nrf_gpio_cfg_output(POWER_EN_PIN);
  nrf_gpio_cfg_output(GPOUT_PIN);
  nrf_gpio_cfg_output(UART_TX_PIN);
  nrf_gpio_cfg_output(I2C_SDA_PIN);
  nrf_gpio_cfg_output(I2C_SCL_PIN);
  nrf_gpio_cfg_output(UART_RTS_PIN);
  nrf_gpio_cfg_output(JACK_EN_PIN);

  // Digital Inputs
  nrf_gpio_cfg_input(SPI_MISO_PIN, NRF_GPIO_PIN_PULLDOWN);
  nrf_gpio_cfg_input(RESET_PIN, NRF_GPIO_PIN_PULLUP);
  nrf_gpio_cfg_input(A_PIN, NRF_GPIO_PIN_PULLDOWN);
  nrf_gpio_cfg_input(B_PIN, NRF_GPIO_PIN_PULLDOWN);
  nrf_gpio_cfg_input(SW_PIN, NRF_GPIO_PIN_PULLDOWN);
  nrf_gpio_cfg_input(GPOUT_PIN, NRF_GPIO_PIN_PULLUP);
  nrf_gpio_cfg_input(UART_RX_PIN, NRF_GPIO_PIN_PULLUP);
  nrf_gpio_cfg_input(UART_CTS_PIN, NRF_GPIO_PIN_PULLUP);
  nrf_gpio_cfg_input(AIN7_PIN, NRF_GPIO_PIN_NOPULL);
  nrf_gpio_cfg_input(AIN6_PIN, NRF_GPIO_PIN_NOPULL);
  nrf_gpio_cfg_input(AIN5_PIN, NRF_GPIO_PIN_NOPULL);
  nrf_gpio_cfg_input(AIN4_PIN, NRF_GPIO_PIN_NOPULL);
  nrf_gpio_cfg_input(AIN1_PIN, NRF_GPIO_PIN_NOPULL);
  nrf_gpio_cfg_input(AIN0_PIN, NRF_GPIO_PIN_NOPULL);
  nrf_gpio_cfg_input(SW1_PIN, NRF_GPIO_PIN_PULLDOWN);
  nrf_gpio_cfg_input(SW2_PIN, NRF_GPIO_PIN_PULLDOWN);
  nrf_gpio_cfg_input(SW3_PIN, NRF_GPIO_PIN_PULLDOWN);

  // configure gpiote with the task to toggle the bit. Takes the pin to toggle and the task config
  nrfx_gpiote_out_config_t config = NRFX_GPIOTE_CONFIG_OUT_TASK_TOGGLE(true);
  err_code = nrfx_gpiote_out_init(SPI_LDAC_PIN, &config);
  NFRX_APP_ERROR_CHECK(err_code);
}

void timer2_event_handler(nrf_timer_event_t event_type, void *p_context)
{
  nrfx_err_t err_code = NRFX_SUCCESS;
  volatile uint32_t uiTXD_PTR;
  volatile uint32_t uiOverflow;
  printf("PTR: %x\n", spim3.p_reg->TXD.PTR);

  // NOTES: Consider connecting WriterListTransmitting to a GPIO line to test that it toggles evenly at 16KHz
  if (WriterListTransmitting==0)
  {
    // As a test indicator, clear the KG_PIN
    // TODO remove this after testing is complete
    nrf_gpio_pin_clear(KG_PIN); // P1.02 = SIO_34

    // Switch the SPIM3 transmission to the other ArrayList, at a starting offset 
    // relative to the value read from TXD.PTR

    // Read the TXD.PTR value from the currently transmitting ArrayList to see 
    // where in the buffer the transmit process has currently reached
    uiTXD_PTR = spim3.p_reg->TXD.PTR;

    // uiOverflow is the number of bytes past the 20ms point into WriterList0 that the TXD.PTR has reached
    uiOverflow = (uiTXD_PTR - (uint32_t)WriterList0) - 1280; // 1280 bytes is (20ms * 16KHz * 4 bytes)

    // Modify the TXD.PTR to point to the beginning of the new array list, plus the overflow amount
    spim3.p_reg->TXD.PTR = (uint32_t)WriterList1 + uiOverflow;

    WriterListTransmitting=1;

    // TODO: Write new audio data to the relevant sections of the two 30ms swing-buffer(s)
  }
  else
  {
    // As a test indicator, set the KG_PIN
    // TODO remove this after testing is complete
    nrf_gpio_pin_set(KG_PIN); // P1.02 = SIO_34

    // Switch the SPIM3 transmission to the other ArrayList, at a starting offset 
    // relative to the value read from TXD.PTR

    // Read the TXD.PTR value from the currently transmitting ArrayList to see 
    // where in the buffer the transmit process has currently reached
    uiTXD_PTR = spim3.p_reg->TXD.PTR;

    // uiOverflow is the number of bytes past the 20ms point into WriterList0 that the TXD.PTR has reached
    uiOverflow = (uiTXD_PTR - (uint32_t)WriterList1) - 1280; // 1280 bytes is (20ms * 16KHz * 4 bytes)

    // Modify the TXD.PTR to point to the beginning of the new array list, plus the overflow amount
    spim3.p_reg->TXD.PTR = (uint32_t)WriterList0 + uiOverflow;

    WriterListTransmitting=0;

    // TODO: Write new audio data to the relevant sections of the two 30ms swing-buffer(s)
  }
}

static void TIMERs_init(void)
{
  // Configure and initialise Timer 0 as a 16-bit 4MHz counter.
  nrfx_err_t err_code = NRFX_SUCCESS;
  nrfx_timer_config_t timer_cfg = NRFX_TIMER_DEFAULT_CONFIG;
  timer_cfg.frequency = NRF_TIMER_FREQ_4MHz;
  timer_cfg.mode      = NRF_TIMER_MODE_TIMER;
  timer_cfg.bit_width = NRF_TIMER_BIT_WIDTH_16;
  err_code = nrfx_timer_init(&m_timer1, &timer_cfg, NULL);
  NFRX_APP_ERROR_CHECK(err_code);

  // Allocate the four PPI channels
  err_code = nrfx_ppi_channel_alloc(&ppi_channel0);
  NFRX_APP_ERROR_CHECK(err_code);
  err_code = nrfx_ppi_channel_alloc(&ppi_channel1);
  NFRX_APP_ERROR_CHECK(err_code);
  err_code = nrfx_ppi_channel_alloc(&ppi_channel2);
  NFRX_APP_ERROR_CHECK(err_code);
  err_code = nrfx_ppi_channel_alloc(&ppi_channel3);
  NFRX_APP_ERROR_CHECK(err_code);

  // set TIMER1 COMPARE[0] count to be 15 (so it generates an event every time the Timer starts or is reset)
  nrfx_timer_compare(&m_timer1,
                     NRF_TIMER_CC_CHANNEL0,
                     15,
                     false);

  // set TIMER1 COMPARE[1] count to be 140 (repeats at 32KHz = 31.25uS)
  nrfx_timer_compare(&m_timer1,
                     NRF_TIMER_CC_CHANNEL1,
                     140,
                     false);

  // Set TIMER1 COMPARE[2] count to be 246 (0.5uS before the COMPARE[2] event)
  nrfx_timer_compare(&m_timer1,
                     NRF_TIMER_CC_CHANNEL2,
                     244,
                     false);

  uint32_t compare0_timer_event_address = nrfx_timer_event_address_get(&m_timer1, NRF_TIMER_EVENT_COMPARE0);
  uint32_t compare1_timer_event_address = nrfx_timer_event_address_get(&m_timer1, NRF_TIMER_EVENT_COMPARE1);
  uint32_t compare2_timer_event_address = nrfx_timer_event_address_get(&m_timer1, NRF_TIMER_EVENT_COMPARE2);
  uint32_t compare3_timer_event_address = nrfx_timer_event_address_get(&m_timer1, NRF_TIMER_EVENT_COMPARE3);

  uint32_t spim_start_task = nrfx_spim_start_task_get(&spim3);
  uint32_t gpiote_clr_task = nrfx_gpiote_clr_task_addr_get(SPI_LDAC_PIN);
  uint32_t gpiote_set_task = nrfx_gpiote_set_task_addr_get(SPI_LDAC_PIN);

  // set TIMER1 COMPARE[3] count to be 250 (repeats at 16KHz = 62.5uS)
  // SHORT the COMPARE[3] event and CLEAR task.
  nrfx_timer_extended_compare(&m_timer1,
                              NRF_TIMER_CC_CHANNEL3,
                              252, // TODO Find out why this needs to be 252 to get real-world 16KHz
                              NRF_TIMER_SHORT_COMPARE3_CLEAR_MASK,
                              false);

  // Connect COMPARE[0] event to TASKS_START of SPIM3 to start transmission every time the event occurs
  err_code = nrfx_ppi_channel_assign(ppi_channel0, compare0_timer_event_address, spim_start_task);
  NFRX_APP_ERROR_CHECK(err_code);

  // Connect COMPARE[1] event to TASKS_START of SPIM3 to start transmission every time the event occurs
  err_code = nrfx_ppi_channel_assign(ppi_channel1, compare1_timer_event_address, spim_start_task);
  NFRX_APP_ERROR_CHECK(err_code);

  // Connect COMPARE[2] event to GPIOTE->TASKS_CLR for the /LDAC signal to make it go LOW
  err_code = nrfx_ppi_channel_assign(ppi_channel2, compare2_timer_event_address, gpiote_clr_task);
  NFRX_APP_ERROR_CHECK(err_code);

  // Connect COMPARE[3] event to GPIOTE->TASKS_CLR for the /LDAC signal to make it go HIGH
  err_code = nrfx_ppi_channel_assign(ppi_channel3, compare3_timer_event_address, gpiote_set_task);
  NFRX_APP_ERROR_CHECK(err_code);

  // Enable the GPIOTE task events for SPI_LDAC_PIN
  nrfx_gpiote_out_task_enable(SPI_LDAC_PIN);

  // Enable the PPI channels
  /*err_code = nrfx_ppi_channel_enable(ppi_channel0);
  NFRX_APP_ERROR_CHECK(err_code);
  err_code = nrfx_ppi_channel_enable(ppi_channel1);
  NFRX_APP_ERROR_CHECK(err_code);
  err_code = nrfx_ppi_channel_enable(ppi_channel2);
  NFRX_APP_ERROR_CHECK(err_code);
  err_code = nrfx_ppi_channel_enable(ppi_channel3);
  NFRX_APP_ERROR_CHECK(err_code);
*/

  // Configure TIMER2 in COUNTER mode, 
  // PPI.FORK COMPARE[3] to COUNT input of TIMER2, to cause TIMER2 to increment on each COMPARE[2] event 
  // Set TIMER2 COMPARE[0] count to be 319
  // When the count reaches 320, we have sent 20ms of samples (2 lots of 2 byte tramsissions)
  // Connect TIMER2 COMPARE[0] event to an Interrupt

   // Configure and initialise Timer 2 in COUNTER mode
  timer_cfg.frequency = NRF_TIMER_FREQ_4MHz;
  timer_cfg.mode      = NRF_TIMER_MODE_COUNTER;
  timer_cfg.bit_width = NRF_TIMER_BIT_WIDTH_16;
  err_code = nrfx_timer_init(&m_timer2, &timer_cfg, timer2_event_handler);
  NFRX_APP_ERROR_CHECK(err_code);

  // set TIMER2 COMPARE[0] count to be 320 (20ms of 2x 16-bit at 16KHz)
  // SHORT the TIMER2 COMPARE[0] event and CLEAR task.
  nrfx_timer_extended_compare(&m_timer2,
                              NRF_TIMER_CC_CHANNEL0,
                              320,
                              NRF_TIMER_SHORT_COMPARE0_CLEAR_MASK,
                              true);

  // Fork Timer 1's COMPARE[3] event to Timer 2's count input task
  uint32_t timer_count_task_addr = nrfx_timer_task_address_get(&m_timer2, NRF_TIMER_TASK_COUNT);
  m_timer2.p_reg->TASKS_COUNT=1;
  
  err_code = nrfx_ppi_channel_fork_assign(ppi_channel3, timer_count_task_addr);
  NFRX_APP_ERROR_CHECK(err_code);

  // Clear Timer 2
  nrfx_timer_clear(&m_timer2);
  // Enable Timer 2
  nrfx_timer_enable(&m_timer2);


  // Clear Timer 1
  nrfx_timer_clear(&m_timer1);
  // Enable Timer 1
  //nrfx_timer_enable(&m_timer1);
}

// Configure the SPIM3, PPI and GPIOTE to enable regular writing to the stereo DAC
void SPIM3_init(void)
{
  nrfx_err_t err_code = NRFX_SUCCESS;

  nrfx_spim_config_t spi_config = NRFX_SPIM_DEFAULT_CONFIG;
  spi_config.ss_pin         = SPI_CSN_PIN;
  spi_config.miso_pin       = NRFX_SPIM_PIN_NOT_USED;
  spi_config.mosi_pin       = SPI_MOSI_PIN;
  spi_config.sck_pin        = SPI_SCK_PIN;
  spi_config.dcx_pin        = NRFX_SPIM_PIN_NOT_USED;
  spi_config.use_hw_ss      = true;
  spi_config.ss_active_high = false;
  spi_config.frequency      = NRF_SPIM_FREQ_1M;
  spi_config.mode           = NRF_SPIM_MODE_0;
  spi_config.bit_order      = NRF_SPIM_BIT_ORDER_MSB_FIRST;

  err_code = nrfx_spim_init(&spim3, &spi_config, NULL, NULL);
  NFRX_APP_ERROR_CHECK(err_code);

  // Enable the use of EasyDMA ArrayList
  spim3.p_reg->TXD.LIST = 1;

  // Function for initializing the SPI master driver instance.
  nrfx_spim_xfer_desc_t xfer_desc = NRFX_SPIM_XFER_TX(WriterList0, BUFFER_SIZE);

  // Function for starting the SPI data transfer
  // NRFX_SPIM_FLAG_REPEATED_XFER = Flag indicating that the transfer will be executed multiple times. 
  // NRFX_SPIM_FLAG_TX_POSTINC = TX buffer address incremented after transfer. 
  // NRFX_SPIM_FLAG_HOLD_XFER = Set up the transfer but do not start it. 
  err_code = nrfx_spim_xfer(&spim3, &xfer_desc, NRFX_SPIM_FLAG_TX_POSTINC|NRFX_SPIM_FLAG_REPEATED_XFER|NRFX_SPIM_FLAG_HOLD_XFER);
  NFRX_APP_ERROR_CHECK(err_code);
}

//********************************************************************
//
//       main()
//
//  Function description
//   Application entry point.
//

void main(void)
{
  unsigned short i;
  unsigned short s;
  float m_fSin;
  float m_fCos;
  float m_fAngle;

  // Read the HWID field of the CONFIGID register in FICR to determine the IC's version
  // number, and therefore apply the correct errata work-arounds.

  // Write some initial data to the SPI write buffers.
  // There are 480 pairs of 16-bit values
  for(i=0;i<480;i++)
  {
    m_fAngle = ((float)i) * AngleStep;    // 640 steps gives 2*PI radians
    m_fSin = (sinf(m_fAngle) + 1) * 2047; // Gain gives full 12-bit range
    m_fCos = (cosf(m_fAngle) + 1) * 2047; // Gain gives full 12-bit range

    // Channel A (Even addresses)
    s = DAC_A | DAC_GAx1 | DAC_RUN |0xAAA;
    WriterList0[i*2].buffer[0] = s >> 8;   // MSB
    WriterList0[i*2].buffer[1] = s & 0xFF; // LSB
    WriterList1[i*2].buffer[0] = s >> 8;   // MSB
    WriterList1[i*2].buffer[1] = s & 0xFF; // LSB
    // Channel B (Odd Addresses)
    s = DAC_B | DAC_GAx1 | DAC_RUN | 0x555;
    WriterList0[(i*2)+1].buffer[0] = s >> 8;   // MSB
    WriterList0[(i*2)+1].buffer[1] = s & 0xFF; // LSB
    WriterList1[(i*2)+1].buffer[0] = s >> 8;   // MSB
    WriterList1[(i*2)+1].buffer[1] = s & 0xFF; // LSB

    /*
    // Channel A (Even addresses)
    s = DAC_A | DAC_GAx1 | DAC_RUN | (unsigned short)lroundf(m_fSin);
    WriterList0[i*2].buffer[0] = s >> 8;   // MSB
    WriterList0[i*2].buffer[1] = s & 0xFF; // LSB
    WriterList1[i*2].buffer[0] = s >> 8;   // MSB
    WriterList1[i*2].buffer[1] = s & 0xFF; // LSB
    // Channel B (Odd Addresses)
    s = DAC_B | DAC_GAx1 | DAC_RUN | (unsigned short)lroundf(m_fCos);
    WriterList0[(i*2)+1].buffer[0] = s >> 8;   // MSB
    WriterList0[(i*2)+1].buffer[1] = s & 0xFF; // LSB
    WriterList1[(i*2)+1].buffer[0] = s >> 8;   // MSB
    WriterList1[(i*2)+1].buffer[1] = s & 0xFF; // LSB
    */
  }

  // Configure the IO pins and their directions
  GPIO_init();
  //SPIM3_init();
  TIMERs_init();
  while(1)
  {
    m_timer2.p_reg->TASKS_COUNT = 1;
    nrf_delay_ms(10);
  }
}

/*************************** End of file ****************************/